/* Regulatory Signal Search (RSS) */
/* Functions: AnalyzeArgs */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rss01.h"

int AnalyzeArgs(int argc, char **argv) {
/* Parses arguments of the main program and sets global variables */
/* Also opens/creates input data and result files */
int rc=OK;
int i, k, iinfile=0;
char mode[4] = "w+t";      /* re-write output file by default */
#ifdef RSSDEBUG
   printf("Actual arguments:\n");
   printf("0: \"%s\"   [v.%s]\n", argv[0], VERSION);
   for(i=1; i<argc; i++)
      printf("%d: \"%s\"\n", i, argv[i]);
   printf("\n");
#endif
   for(i=1; i<argc; i++) {
      if(argv[i][0]=='-' || argv[i][0]=='/')     /* Option found */
         switch(argv[i][1]) {
            case 'A':
            case 'a':            /* Append to output file */
               mode[0] = 'a';
               break;
            case 'B':
            case 'b':            /* Abs limit of non-improving iterations */
               k = atoi(argv[i]+2);
               if(k<1)              rc |= ERARG;
               else                 afterabs = k;
               break;
            case 'C':
            case 'c':            /* Console output level */
               k = atoi(argv[i]+2);
               if(k<0 || k>2)       rc |= ERARG;
               else                 cdef = k;
               break;
            case 'D':
            case 'd':            /* Write debug information into output file */
               k = atoi(argv[i]+2);
               if(k<0 || k>3)       rc |= ERARG;
               else                 debug = k;
               break;
            case 'G':
            case 'g':            /* Mode of aggregation */
               if(argv[i][2]=='s' || argv[i][2]=='S')       k = AG_SUM;
               else if(argv[i][2]=='m' || argv[i][2]=='M')  k = AG_MAX;
               else k = atoi(argv[i]+2);
               if(k!=AG_SUM && k!=AG_MAX) rc |= ERARG;
               else                 aggr = k;
               break;
            case '?':
            case 'H':
            case 'h':            /* Get help on arguments */
               printf("Usage: rss01 [Options] Infile [Outfile]\n\n");
               printf("Infile:  Name of input FASTA file\n");
               printf("Outfile: Name for result data file (*.out default)\n\n");
               printf("Options (case insensitive):\n");
               printf(" -A    Append to output file (overwrite default)\n");
               printf(" -Bxx  Absolute limit of non-improving iterations (%d default)\n", afterabs);
               printf(" -Cxx  Console output level xx=0..2 (%d default)\n", CDEF);
               printf(" -Dxx  Write level xx=0..3 diagnostics into output file (%d default)\n", DG_DEF );
               printf(" -Gx   Aggregation mode: %d|S=sum, %d|M=max (%d default)\n", AG_SUM, AG_MAX, AG_DEF);
               printf(" -H    Print help information\n");
               printf(" -Ixx  Limit of non-improving iterations by profile (%d default)\n", ITER_DEF);
               printf(" -Lxx  Set signal word length xx (%d default)\n", LDEF);
               printf(" -Mx   Permutation: %d=1 %d|N=n %d|F=full(n!) %d|R=random %d|A=active (%d default)\n",
                  PM_IDENT, PM_CYCLE, PM_FULL, PM_RANDOM, PM_ACTIVE, PM_DEF);
               printf(" -Pxx  Min pairs per sequence to try (0 means all pairs, n/2 default)\n");
               printf(" -Rxx  Number of best results to print (%d default)\n", RDEF);
               printf(" -Sxx  Seed value for random number generator (randomize by default)\n");
               printf(" -Txx  Set quality threshold (L/2 default)\n");
               printf(" -Uxx  Limit number of permutations (%d default)\n", LIM_DEF);
               printf(" -Vxx  Limit of new permutations at a time (%d default)\n", PV_DEF);
               printf(" -Xn   Proximity: 1=fill matrix, 0=calculate on the fly (default 1)\n\n");
               rc |= HELP;
               break;
            case 'I':
            case 'i':            /* Max number of non-improving iterations */
               k = atoi(argv[i]+2);
               if(k<1)              rc |= ERARG;
               else                 aftermax = k;
               break;
            case 'L':
            case 'l':            /* Set length of a word */
               k = atoi(argv[i]+2);
               if(k<1 || k>LMAX)    rc |= ERARG;
               else                 l = k;
               break;
            case 'M':
            case 'm':            /* Mode of permutation */
               if(argv[i][2]=='n' || argv[i][2]=='N')       k = PM_CYCLE;
               else if(argv[i][2]=='f' || argv[i][2]=='F')  k = PM_FULL;
               else if(argv[i][2]=='r' || argv[i][2]=='R')  k = PM_RANDOM;
               else if(argv[i][2]=='a' || argv[i][2]=='A')  k = PM_ACTIVE;
               else k = atoi(argv[i]+2);
               if(k<PM_IDENT || k>PM_ACTIVE) rc |= ERARG;
               else                 pmode = k;
               break;
            case 'P':
            case 'p':            /* Min pairs per sequence to try (0=all) */
               k = atoi(argv[i]+2);
               pairs = k;
               break;
            case 'R':
            case 'r':				/* Set number of best results to print */
            	k = atoi(argv[i]+2);
               if(k<1)					rc |= ERARG;
               else						rdef = k;
               break;
            case 'S':
            case 's':				/* Set seed value for random generator */
            	seed = (unsigned)atol(argv[i]+2);
               break;
            case 'T':
            case 't':            /* Set threshold for quality */
               k = atoi(argv[i]+2);
               if(k<1)              rc |= ERARG;
               else                 threshold = k;
               break;
            case 'U':
            case 'u':            /* Limit number of permutations */
               k = atoi(argv[i]+2);
               if(k<1)              rc |= ERARG;
               else                 plimit = k;
               break;
            case 'V':
            case 'v':            /* Limit of new permutations at a time */
               k = atoi(argv[i]+2);
               if(k<1 || k>MAXPVECT)   rc |= ERARG;
               else                 pvmax = k;
               break;
            case 'X':
            case 'x':            /* Use of proximity matrix */
               if(argv[i][2]=='0')        matrix = 0;
               else if(argv[i][2]=='1')   matrix = 1;
               else                 rc |= ERARG;
               break;
            default:                rc |= ERARG;
         }
      else {                  /* Filename assumed */
         if(!infile && iinfile==0) {
            if((infile = fopen(argv[i], "rt"))==NULL)
               rc |= (ERARG | ERDATA | ERIO);
            iinfile = i;
         }
         else if(!outfile) {
            if((outfile = fopen(argv[i], mode))==NULL)
               rc |= (ERARG | ERIO);
            else {
               strncpy(outfilename, argv[i], MAXPATH);
               outfilename[MAXPATH]=0;
            }
         }
         else
               rc |= ERARG;
      }
      if(rc & ERARG) {
         fprintf(stderr, "Invalid argument \"%s\"\n", argv[i]);
         break;
      }
   }
   if(rc & HELP)  return rc;
   if(!infile) {
      rc |= (ERARG | ERDATA);
      fprintf(stderr, "Cannot open input data file \"%s\"\n", argv[iinfile]);
   }
   else if(!outfile) {
      strncpy(outfilename, argv[iinfile], MAXPATH);
      strcat(outfilename, DEFOUT);
      if((outfile = fopen(outfilename, mode))==NULL) {
         rc |= (ERARG | ERIO);
         fprintf(stderr, "Cannot open output data file \"%s\"\n", outfilename);
      }
   }
   if(debug>1) {
      fprintf(outfile, "Actual arguments:\n");
      fprintf(outfile, "0: \"%s\"   [v.%s]\n", argv[0], VERSION);
      for(i=1; i<argc; i++)
         fprintf(outfile, "%d: \"%s\"\n", i, argv[i]);
      fprintf(outfile, "\n");
   }
   if(l==EMPTY)            l = LDEF;
   if(threshold==EMPTY)    threshold = l/2;
   return rc;
}

