/* Regulatory Signal Search (RSS) */
/* Functions: Construction */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include "rss01.h"

int Construction(SBORKA *sborka) {
/* Builds a regulator signal system for current arrangement */
/* For the moment only reproduces input data and a tree */
int rc=OK;
int i, j, k, r, z, w, *c;
short *s, *q, *smax=NULL;
int result[NMAX], qual[NMAX], qt, nonzero, avg;
   for(i=0; i<NMAX; i++)
      { result[i] = EMPTY;  qual[i] = EMPTY; }
   for(z=logn-1; z>0; z--) {    /* Main loop of assembly */
		if(cdef>1)
      	printf("Level %d\n", z);
      if(z==logn-1)  rc |= MakeBottomLevel(sborka);
      else           rc |= MakeNextLevel(sborka, z);
      if(rc!=OK)  break;
      if(debug>2) {             /* Write subsystems/costs for next level */
         fprintf(outfile, "\n i  j   (System for next level)    "
            "(Costs for the next level)\n");
         for(i=0,s=sborka->suu,c=sborka->cuu; i<mwz; i++) {
            for(j=0; j<mwz; j++) {
               if(i<mw)  fprintf(outfile, "%02d,", i);
               else fprintf(outfile, "ZZ,");
               if(j<mw)  fprintf(outfile, "%02d:", j);
               else fprintf(outfile, "ZZ:");
               for(k=0; k<nn; k++,s++)
                  if(sborka->tree[z-1][k] >=0)
                     if(*s==ZERO) fprintf(outfile, " ZZ");
                     else fprintf(outfile, " %2d", *s);
                  else  fprintf(outfile, "  *");
               fprintf(outfile, "     (");
               w = 1<<(z-1);
               for(k=0; k<w; k++,c++)
                  fprintf(outfile, "%5d ", *c);
               fprintf(outfile, ")\n");
            }
            fprintf(outfile, "\n");
         }
         fprintf(outfile, "Dz=%1d: ", z-1);
         for(k=0; k<nn; k++)
            if((w=sborka->tree[z-1][k])>=0)    fprintf(outfile, " %2d", w);
            else  fprintf(outfile, "  *");
         fprintf(outfile, "\n");
         if(fflush(outfile)!=0)  return ERIO;
      }
   }
                                       /* Work with final system (z=0) */
   c=sborka->cuu; s=sborka->suu; w=-1;
   for(i=0; i<mwz; i++)       /* Find optimal pair of positions for <u1,u2> */
      for(j=0; j<mwz; j++,c++,s+=nn)
         if(*c > w)  { w = *c;  smax = s; }
                              /* Calculate quality (temporary method) */
   for(i=0,s=smax; i<n; i++,s++) {
      j = sborka->p[i];
      result[j] = *s;
   }
   for(i=0,qt=0,nonzero=n; i<n; i++) {
      k = sborka->p[i];
      qual[k] = 0;
      if(result[k]==mw)
         { nonzero--; continue; }
      for(j=0; j<n; j++) {
         r = sborka->p[j];
         if(k==r || result[r]==mw)  continue;
         else qual[k] += Proximity(k, r, result[k], result[r]);
      }
      qt += qual[k];
   }
   avg = nonzero>1 ? (qt*100)/(nonzero*(nonzero-1)) : 0;
   qt = nonzero>1 ? (qt*100)/(nonzero-1) : 0;
   if(debug>1) {
      fprintf(outfile, "Signal found (power: %d, quality: %.2f, "
         "avg proximity: %.2f)\n", nonzero, (float)(qt/100.), (float)(avg/100.));
      for(i=0; i<NMAX; i++) {
         if(result[i]==mw)
            fprintf(outfile, "%3d:%.*s\n", i+1, maxname, name[i]);
         else if(result[i]==EMPTY)
            continue;
         else fprintf(outfile, "%3d:%.*s[%04d] %.*s (%d, %5.2f)\n",
            i+1, maxname, name[i], result[i]+1, l, seq[i]+result[i], qual[i],
            nonzero>1?((float)qual[i])/(nonzero-1):0.);
      }
      fprintf(outfile, "\n");
      if(fflush(outfile)!=0)  return ERIO;
   }
	if(cdef) {
      printf("P[%d]: ", first-1);
      for(i=0,q=sborka->p; i<n; i++,q++)  printf("%d ", *q+1);
   	printf("=> pow=%d  q=%.2f  avprx=%.2f\n", nonzero,
      	(float)(qt/100.), (float)(avg/100.));
   }
   return rc;
}

