/* Regulatory Signal Search (RSS) */
/* Functions: AnalyzeHistory, TryProfile, TryCoverage, Recurrent */
/*            SortProfile, SortCoverage, SortPairs */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"
                        /* To implement later: removing sequences ! */

int AnalyzeHistory(void) {
/* Aggregates profile and coverage data for processed permutations */
int i;
   for(i=0; i<n0; i++)           /* Initialize order */
      sq[i].seqn = i;
   if(after>afterabs) {          /* Check if quality doesn't improve */
      if(cdef)
      	printf("Terminated due to non-improvement criterion\n");
      if(debug)
      	fprintf(outfile, "Terminated due to non-improvement criterion\n");
      return EMPTY;              /* during last 'aftermax' iterations */
   }
   if((i=TryProfile())!=0 &&     /* Propose candidates based on profile */
      after<aftermax) {        /* if quality tends to improve */
		if(cdef>1)
      	printf("Generate %d permutation(s) based on principle 1\n", i);
      if(debug>1)
         fprintf(outfile, "Generate %d permutation(s) based on principle 1\n", i);
      return OK;                    /* Candidate(s) found */
   }
   for(i=0; i<=pvmax; i++)    pvect[i][0]=EMPTY;
   
   if((i=TryCoverage())==0) {    /* Propose candidates based on coverage */
      if(cdef)
      	printf("Terminated due to both profile and coverage criteria\n");
      if(debug)
      	fprintf(outfile, "Terminated due to both profile and coverage criteria\n");
      return EMPTY;                 /* Candidate(s) not found */
   }
	if(cdef>1)
   	printf("Generate %d permutation(s) based on principle 2\n", i);
   if(debug>1)
      fprintf(outfile, "Generate %d permutation(s) based on principle 2\n", i);
   return OK;
}

int TryProfile(void) {
/* Calculates a profile and proposes candidate(s) returning number of them */
int cnum, i, j, k, i0, i1;
   for(i=0; i<n0; i++)  sq[i].profile=0;     /* clear joint profile and */
   for(i=0; i<n0; i++) {                     /* recalculate it */
      if(lambda[i][0]==EMPTY)
         { sq[i].profile=EMPTY; continue; }
      if(aggr==AG_SUM) {                     /* sum mode */
         for(j=0; j<mw; j++)
            sq[i].profile += lambda[i][j];
      }
      else if(aggr==AG_MAX) {                /* max mode */
         for(j=0; j<mw; j++)
            if(sq[i].profile < lambda[i][j])
               sq[i].profile = lambda[i][j];
      }
      else return 0;                         /* unknown mode */
   }
                                             /* sort and propose */
   qsort((void*)sq, n0, sizeof(sq[0]), SortProfile);
   for(i=0; i<pvmax; i++)
      if(pvect[i][0]==EMPTY)  break;         /* find free row of P-vector */
   for(j=0,k=0; j<n0; j++)                   /* write sorted permutation */
      if(sq[j].profile != EMPTY)
         pvect[i][k++]=(short)sq[j].seqn;
   for( ; k<NMAX; k++)     pvect[i][k]=EMPTY;  /* clear tail */
   i0=i; i1=i+1; cnum=1;
   for(i=0; i<n-1; i++) {                    /* find "synonyms" if any */
      if(sq[i].profile!=sq[i+1].profile)
         continue;
      for( ; i1<pvmax+1; i1++)   /* one spare on case if [i0] will be removed */
         if(pvect[i1][0]==EMPTY)             /* find first free row */
            break;
      if(i1>=pvmax+1) break;              /* ignore remaining opportunities */
      for(j=0; j<NMAX; j++)
         pvect[i1][j]=pvect[i0][j];       /* duplicate a permutation */
      j=pvect[i1][i];                     /* then swap equal elements */
      pvect[i1][i]=pvect[i1][i+1]; pvect[i1][i+1]=(short)j;
#ifdef RSSPREVENT
      if(Recurrent(i1))  pvect[i1][0]=EMPTY;
      else
#endif
         { i1++; cnum++; }
   }
#ifdef RSSPREVENT
   if(Recurrent(i0)) {
      for(i=0; i<NMAX; i++)
         pvect[i0][i]= pvect[pvmax][i];   /*use the spare permutation if any */
      pvect[pvmax][0]=EMPTY;
      cnum--;
   }
#endif
   if(pvect[pvmax][0]!=EMPTY) {           /* delete spare permutation */
      pvect[pvmax][0]=EMPTY;
      cnum--;
   }
   return cnum;
}

int TryCoverage(void) {
/* Calculates a coverage and proposes candidate(s) returning number of them */
int (/*_USERENTRY*/ *fcmp)(const void *, const void *);
int cnum, i, j, k, i0, i1;
   for(i=0; i<n0; i++)                       /* clear joint coverage and */
      { sq[i].coverage=0; sq[i].pairs=0; }
   for(i=0,k=0; i<n0-1; i++)                 /* then recalculate it */
      for(j=i+1; j<n0; j++,k++) {
         if(cover[k]==EMPTY || cover[k]==0)  continue;
         sq[i].pairs++; sq[j].pairs++;
         if(aggr==AG_SUM) {                     /* sum mode */
            sq[i].coverage += cover[k];
            sq[j].coverage += cover[k];
         }
         else if(aggr==AG_MAX) {                /* max mode */
            if(sq[i].coverage < cover[k])
               sq[i].coverage = cover[k];
            if(sq[j].coverage < cover[k])
               sq[j].coverage = cover[k];
         }
         else return 0;                         /* unknown mode */
      }
                                          /* Check termination conditions */
   if(pairs==0) {                            /* all pairs needed */
      for(i=0; i<n0; i++)
         if(sq[i].profile!=EMPTY && sq[i].pairs<(n-1))
            break;
      if(i>=n0) return 0;                    /* all pairs tested */
      fcmp = SortPairs;                      /* otherwise sort by pairing */
   }
   else {                                    /* pair per seq limit is set */
      for(i=0; i<n0; i++)
         if(sq[i].profile!=EMPTY && sq[i].pairs<pairs)
            break;
      if(i>=n0) return 0;                    /* pairing above limit everywhere */
      fcmp = SortCoverage;                   /* otherwise sort by coverage */
   }
                                          /* Otherwise sort and propose */
   qsort((void*)sq, n0, sizeof(sq[0]), fcmp);   /* sort as decided */
   for(i=0; i<pvmax; i++)
      if(pvect[i][0]==EMPTY)  break;         /* find free row of P-vector */
   for(j=0,k=0; j<NMAX; j++)                 /* write sorted permutation */
      if(sq[j].profile != EMPTY)
         pvect[i][k++]=(short)sq[j].seqn;
   for( ; k<NMAX; k++)     pvect[i][k]=EMPTY;  /* clear tail */
   i0=i; i1=i+1; cnum=1;
   for(i=0; i<n-1; i++) {                    /* find "synonyms" if any */
      if((pairs==0 && sq[i].pairs!=sq[i+1].pairs) ||
         (pairs>0 && sq[i].coverage!=sq[i+1].coverage))
         continue;
      for( ; i1<pvmax+1; i1++)   /* one spare on case if [i0] will be removed */
         if(pvect[i1][0]==EMPTY)             /* find first free row */
            break;
      if(i1>=pvmax+1) break;              /* ignore remaining opportunities */
      for(j=0; j<NMAX; j++)
         pvect[i1][j]=pvect[i0][j];       /* duplicate a permutation */
      j=pvect[i1][i];                     /* then swap equal elements */
      pvect[i1][i]=pvect[i1][i+1]; pvect[i1][i+1]=(short)j;
#ifdef RSSPREVENT
      if(Recurrent(i1))  pvect[i1][0]=EMPTY;
      else
#endif
         { i1++; cnum++; }
   }
#ifdef RSSPREVENT
   if(Recurrent(i0)) {
      for(i=0; i<NMAX; i++)
         pvect[i0][i]= pvect[pvmax][i];   /*use the spare permutation if any */
      pvect[pvmax][0]=EMPTY;
      cnum--;
   }
#endif
   if(pvect[pvmax][0]!=EMPTY) {           /* delete spare permutation */
      pvect[pvmax][0]=EMPTY;
      cnum--;
   }
   return cnum;
}

int Recurrent(int k) {
/* Checks if k-th element of P-vector has been processed earlier and */
/* returns YES (recurrence) or NO (new permutation). */
#define YES 1
#define NO  0
HISTORY *h=hist1;
int i, j;
   for(i=0; i<pvmax; i++) {                     /* Check queue first */
      if(i==k || pvect[i][0]==EMPTY) continue;
      for(j=0; j<n; j++)
         if(pvect[i][j] != pvect[k][j]) break;
      if(j>=n)  { recursum++; return YES; }
   }
   while(h) {                                   /* Then check history */
      for(i=0; i<n; i++)
         if(h->p[i] != pvect[k][i]) break;
      if(i>=n)  { recursum++; return YES; }
      h=h->next;
   }
   return NO;
}

int SortProfile(const void *a, const void *b) {
   if     (((SNUM*)a)->profile == EMPTY) return -1;
   else if(((SNUM*)b)->profile == EMPTY) return 1;
   else if(((SNUM*)a)->profile > ((SNUM*)b)->profile) return -1;
   else if(((SNUM*)a)->profile < ((SNUM*)b)->profile) return 1;
   else if(((SNUM*)a)->seqn < ((SNUM*)b)->seqn) return 1;
   else return -1;
}

int SortCoverage(const void *a, const void *b) {
   if     (((SNUM*)a)->profile == EMPTY) return -1;
   else if(((SNUM*)b)->profile == EMPTY) return 1;
   else if(((SNUM*)a)->coverage < ((SNUM*)b)->coverage) return -1;
   else if(((SNUM*)a)->coverage > ((SNUM*)b)->coverage) return 1;
   else if(((SNUM*)a)->seqn < ((SNUM*)b)->seqn) return 1;
   else return -1;
}

int SortPairs(const void *a, const void *b) {
   if     (((SNUM*)a)->profile == EMPTY) return -1;
   else if(((SNUM*)b)->profile == EMPTY) return 1;
   else if(((SNUM*)a)->pairs < ((SNUM*)b)->pairs) return -1;
   else if(((SNUM*)a)->pairs > ((SNUM*)b)->pairs) return 1;
   else if(((SNUM*)a)->seqn < ((SNUM*)b)->seqn) return 1;
   else return -1;
}

