/* Regulatory Signal Search (RSS) */
/* Functions: GenTree */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"

int GenTree(SBORKA *sborka) {
/* Generates a tree for n points based on the permutation p in sborka */
/* The tree is built from initial numbers despite the permutation and */
/* exemption of the sequences may be made. */
int i, j, k, r, d;
int sys_size;
short **tree;
short *q;
   if(!sborka)    return ERMEM;   
   sys_size = mwz*mwz*nn;
   if((sborka->tree = (short**)calloc(LOGNMAX, sizeof(short*)))==NULL ||
      (sborka->suu = (short*)calloc(sys_size, sizeof(short)))==NULL ||
      (sborka->suv = (short*)calloc(sys_size, sizeof(short)))==NULL ||
      (sborka->l1 = (short*)calloc(mwz, sizeof(short)))==NULL ||
      (sborka->l2 = (short*)calloc(mwz, sizeof(short)))==NULL)
      return ERMEM;
   tree = sborka->tree;
   for(i=0; i<LOGNMAX; i++)   tree[i]=NULL;
   for(k=0,r=nn; k<logn; k++,r>>=1) {
      if((tree[k]=(short*)calloc(nn, sizeof(short)))==NULL) {   /* No memory */
         for( --k; k>=0; k--)          /* Free allocated earlier */
            if(tree[k])
               { free(tree[k]); tree[k]=NULL; }
         return ERMEM;
      }
      if(k==0) {              /* 0th level - write source permutation */
         for(i=0; i<n; i++)   tree[k][i] = sborka->p[i];
         for(  ; i<nn; i++)   tree[k][i] = EMPTY;
      }
      else {                  /* fill underlying level from the last one */
         for(i=0,j=0,d=0; i<nn; i+=2,j++,d++) {
            if(d>=r)    { j+=r; d=0; }
            tree[k][j] = tree[k-1][i];
            tree[k][j+r] = tree[k-1][i+1];
         }
      }
   }
                        /* Initialize other parts of sborka */
   for(i=0,q=sborka->suu; i<sys_size; i++,q++)  *q = EMPTY;
   for(i=0,q=sborka->suv; i<sys_size; i++,q++)  *q = EMPTY;
   for(i=0,q=sborka->l1; i<mwz; i++,q++)  *q = EMPTY;
   for(i=0,q=sborka->l2; i<mwz; i++,q++)  *q = EMPTY;
   if(debug>2) {
      fprintf(outfile, "Tree:\n");
      for(k=0; k<logn; k++) {
         fprintf(outfile, "Dz=%1d: ", k);
         for(i=0; i<nn; i++)
            if((j = sborka->tree[k][i])!=EMPTY) fprintf(outfile, "%3.1d", j);
            else                                fprintf(outfile, "  *");
         fprintf(outfile, "\n");
      }
      if(fflush(outfile)!=0)  return ERIO;
   }
	if(cdef>1)
   	printf("G-tree built (%d levels, %d nodes)\n", logn, n);
   return OK;
}

