/* Regulatory Signal Search (RSS) */
/* Functions: AddHistory, BestResult, Profile, Coverage */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"

int AddHistory(SBORKA *sborka) {
/* Appends found system to the end of history with making calculations for it */
int rc=OK, i, j, k, r, *c, w;
short *s, *smax=NULL;
HISTORY *h, *hh;
   if((h = (HISTORY*)malloc(sizeof(HISTORY)))==NULL)  return ERMEM;
   h->n = n;  h->nn = nn;  h->logn = logn;
   h->p = sborka->p;  sborka->p = NULL;
   if((h->s = (int*)calloc(n0, sizeof(int)))==NULL ||
      (h->qw = (int*)calloc(n0, sizeof(int)))==NULL ||
      (h->qwavg = (int*)calloc(n0, sizeof(int)))==NULL)   return ERMEM;
   for(i=0; i<n0; i++)
      { h->s[i]=EMPTY; h->qw[i]=EMPTY; h->qwavg[i]=EMPTY; }
   c=sborka->cuu; s=sborka->suu; w=-1;
   for(i=0; i<mwz; i++)             /* Find optimum section of final system */
      for(j=0; j<mwz; j++,c++,s+=nn)
         if(*c > w)  { w = *c;  smax = s; }
   for(i=0,s=smax; i<n; i++,s++)    /* Store found positions in the history */
      { j = h->p[i];  h->s[j] = *s; }
   h->qt = 0;  h->power = n;
   for(i=0; i<n; i++) {             /* Calculate quality and power */
      k = h->p[i];
      h->qw[k] = 0;
      if(h->s[k]==ZERO)    { h->power--; continue; }
      for(j=0; j<n; j++) {
         r = h->p[j];
         if(k==r || h->s[r]==ZERO)  continue;
         else h->qw[k] += Proximity(k, r, h->s[k], h->s[r]);
      }
      h->qt += h->qw[k];
   }
                                    /* Averaging */
   if((k=h->power)>1)   h->qtavg = (h->qt*100)/(k*(k-1));
   else  h->qtavg = 0;
   h->qt = k>1 ? (h->qt*100)/(k-1) : 0;
   for(i=0; i<n0; i++)
      if(h->qw[i]!=EMPTY)
         h->qwavg[i] = k>1 ? (h->qw[i]*100)/(k-1) : 0;
                                    /* Set global maximum data */
   if(h->qt<=qmaximum)  after++;
   else  { qmaximum=h->qt; after=0; }
												/* Check if the signal is already known */
   for(hh=hist1; hh!=NULL; hh=hh->next) {
   	if(hh->n!=h->n || hh->nn!=h->nn || hh->logn!=h->logn)
      	continue;
   	if(hh->qt!=h->qt || hh->qtavg!=h->qtavg || hh->power!=h->power)
      	continue;
      for(i=0; i<n; i++)
      	if(hh->s[i]!=h->s[i]) break;
      if(i>=n) break;
   }
   h->next = NULL;                  /* Append to history list (semaphoring!) */
   if(hist1==NULL)
      { h->last=NULL; h->pnum=0; hist1=h; hist2=h; }
   else
      { h->last=hist2; h->pnum=hist2->pnum+1; hist2->next=h; hist2=h; }
   rc |= Profile(h);
   rc |= Coverage(sborka);
   if(hh!=NULL) {								/* This signal is already known, */
   	h->qt=h->qtavg=0;						/* prevent its multiple selection */
   }
   return rc;
}

int BestResult(int resnum) {
/* Looks for the best results by quality and by avg proximity in the history */
HISTORY *h, *hq, *ha;
int i;
   for(h=hist1,hq=NULL,ha=NULL; h!=NULL; h=h->next) {
      if(hq==NULL || h->qt>hq->qt || (h->qt==hq->qt && h->qtavg>hq->qtavg))
         hq = h;
      if(ha==NULL || h->qtavg>ha->qtavg || (h->qtavg==ha->qtavg && h->qt>ha->qt))
         ha = h;
   }
   if(hq==NULL || ha==NULL)  return ERINT;
   fprintf(outfile, "Best quality result #%d [%d/%d] (power: %d, quality: %.2f, "
      "avg proximity: %.2f)\n", resnum, hq->pnum, first, hq->power,
      (float)(hq->qt/100.), (float)(hq->qtavg/100.));
   if(debug>1) {
   	fprintf(outfile, "P[%d]: ", hq->pnum);
      for(i=0; i<n; i++)
         if(hq->p[i]==EMPTY)  fprintf(outfile, " * ");
         else fprintf(outfile, "%d ", hq->p[i]+1);
      fprintf(outfile, "\n");
   }
   for(i=0; i<n0; i++) {
      if(hq->s[i]==mw)
         fprintf(outfile, "%3d:%.*s\n", i+1, maxname, name[i]);
      else if(hq->s[i]==EMPTY)
         continue;
      else fprintf(outfile, "%3d:%.*s[%04d] %.*s (%d, %5.2f)\n", i+1, maxname,
      	name[i], hq->s[i]+1, l, seq[i]+hq->s[i], hq->qw[i],
         (float)(hq->qwavg[i]/100.));
   }
   fprintf(outfile, "\n");
   printf("Best quality result #%d [%d/%d] (power: %d, quality: %.2f, avg "
      "proximity: %.2f)\n", resnum, hq->pnum, first, hq->power,
      (float)(hq->qt/100.), (float)(hq->qtavg/100.));
	hq->qt=0; hq->qtavg=0;			/* to exclude this result next time */
   if(ha!=hq) {
      fprintf(outfile, "Best avg proximity result #%d [%d/%d] (power: %d, quality"
         ": %.2f, avg proximity: %.2f)\n", resnum, ha->pnum, first, ha->power,
         (float)(ha->qt/100.), (float)(ha->qtavg/100.));
      if(debug>1) {
         fprintf(outfile, "P[%d]: ", ha->pnum);
         for(i=0; i<n; i++)
            if(ha->p[i]==EMPTY)  fprintf(outfile, " * ");
            else fprintf(outfile, "%d ", ha->p[i]+1);
         fprintf(outfile, "\n");
      }
      for(i=0; i<n0; i++) {
         if(ha->s[i]==mw)
            fprintf(outfile, "%3d:%.*s\n", i+1, maxname, name[i]);
         else if(ha->s[i]==EMPTY)
            continue;
         else fprintf(outfile, "%3d:%.*s[%04d] %.*s (%d, %5.2f)\n", i+1, maxname,
            name[i], ha->s[i]+1, l, seq[i]+ha->s[i], ha->qw[i],
            (float)(ha->qwavg[i]/100.));
      }
      fprintf(outfile, "\n");
      printf("Best avg proximity result #%d [%d/%d] (power: %d, quality: %.2f, "
         "avg proximity: %.2f)\n", resnum, ha->pnum, first, ha->power,
         (float)(ha->qt/100.), (float)(ha->qtavg/100.));
		ha->qt=0; ha->qtavg=0;			/* to exclude this result next time */
   }
   if(fflush(outfile)!=0)  return ERIO;
   return OK;
}

int Profile(HISTORY *h) {
/* Modifies matrix lambda with results of the history update */
int rc=OK, i, j, k;
   if(h==NULL)  return ERINT;
   for(i=0; i<n0; i++)
      if((j=h->s[i])!=ZERO && j!=EMPTY)
         lambda[i][j] += h->qw[i];
      else if(j==EMPTY)    /* clear whole row */
         for(k=0; k<mw; k++)
            lambda[i][k]=EMPTY;
   return rc;
}

int Coverage(SBORKA *sborka) {
/* Adds coverage of new built system to 'cover' matrix */
int i, j, k, r, z, nnp, nnp2, d;
short *p;
   d = logn0-logn+1; /* quant of coverage for bottom level of the tree */
   for(z=logn-1,nnp=2,nnp2=4; z>0; z--,nnp=nnp2,nnp2*=2,d++) {
      if((p=sborka->tree[z])==NULL)    return ERINT;
      for(k=0; k<nn; k+=nnp2,p+=nnp2) {
         i = p[0];  j = p[nnp];
         if(i==j || i==EMPTY || j==EMPTY)   return ERINT;
         if(i>j) { r=i; i=j; j=r; }    /* swap if needed */
         r = i*(n0-1) - (i*(i+1))/2 + j-1;   /* i,j -> r (in triangle matrix) */
         cover[r] += d;
      }
   }
   return OK;
}

