/* Regulatory Signal Search (RSS) */
/* Functions: MakeBottomLevel */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"

int MakeBottomLevel(SBORKA *sborka) {
/* Builds suu[] system for bottom (logn-1)-th level of G-tree */
/* Allocates and fills fuu[], cuu[] for that level */
int rc=OK, nn4=nn/4;
int fc_size, part1, node, u1, v1, u2, v2;
int i, j, k, r, k1, k2, t1, t2, maxpos1=-1, maxpos2=-1, maxsum, psum1, psum2;
int fu1u2, fu1v1, fu1v2, fu2v1, fu2v2, fv1v2, *cuuij;
short *suuij, *q1, *q2;
   if(!sborka)    return ERMEM;
   fc_size = mwz * mwz * nn4;
   if((sborka->cuu = (int*)calloc(fc_size, sizeof(int)))==NULL)
      return ERMEM;
   for(k=0,cuuij=sborka->cuu; k<fc_size; k++,cuuij++)  *cuuij=EMPTY;
   sborka->cuv = NULL;
                                       /* Loop by partitions */
   for(part1=0,node=0; node<nn; part1++,node+=4) {
      u1 = sborka->tree[logn-1][node];
      v1 = sborka->tree[logn-1][node+1];
      u2 = sborka->tree[logn-1][node+2];
      v2 = sborka->tree[logn-1][node+3];
      cuuij = sborka->cuu + part1;  /* c[i][j] (0,0) for current partition */
      suuij = sborka->suu + node;   /* s[i][j] (0,0) of the part. system */

      if(v1==EMPTY && v2==EMPTY) {        /* Case 1: 1--1 */
         for(i=0; i<mwz; i++)
            for(j=0; j<mwz; j++) {
               if((*cuuij = Proximity(u1, u2, i, j))>=threshold) {
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = ZERO;           suuij[3] = ZERO;
               }
               else {
                  *cuuij = 0;
                  suuij[0] = ZERO;           suuij[1] = ZERO;
                  suuij[2] = ZERO;           suuij[3] = ZERO;
               }
               cuuij += nn4;  suuij += nn;      /* next j */
            }
      }
      else if(v2==EMPTY) {                /* Case 2a: 2--1 */
         for(i=0; i<mwz; i++)
            for(j=0; j<mwz; j++) {
               fu1u2 = Proximity(u1, u2, i, j);
               maxsum = -1;
               for(k=0,k1=0; k<mwz; k++) {   /* Find max on (virtual) L1 */
                  fu1v1 = Proximity(u1, v1, i, k);
                  fu2v1 = Proximity(u2, v1, j, k);
                  if((i!=mw && j!=mw && k!=mw && (fu1u2+fu1v1+fu2v1)/3>=threshold) ||
                     (i!=mw && j!=mw && k==mw && fu1u2>=threshold) ||
                     (i!=mw && j==mw && k!=mw && fu1v1>=threshold) ||
                     (i==mw && j!=mw && k!=mw && fu2v1>=threshold))
                     if((fu1v1+fu2v1) > maxsum)
                        { maxsum = fu1v1+fu2v1; maxpos1 = k; k1++; }
               }
               if(k1==0) {                   /* (virtual) L1 is empty */
                  *cuuij = 0;
                  suuij[0] = ZERO;           suuij[1] = ZERO;
                  suuij[2] = ZERO;           suuij[3] = ZERO;
               }
               else {                        /* take max on (virtual) L1 */
                  *cuuij = fu1u2 + maxsum;
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = (short)maxpos1; suuij[3] = ZERO;
               }
               cuuij += nn4;  suuij += nn;      /* next j */
            }
      }
      else if(v1==EMPTY) {                /* Case 2b: 1--2 */
         for(i=0; i<mwz; i++)
            for(j=0; j<mwz; j++) {
               fu1u2 = Proximity(u1, u2, i, j);
               maxsum = -1;
               for(r=0,k2=0; r<mwz; r++) {   /* Find max on (virtual) L2 */
                  fu1v2 = Proximity(u1, v2, i, r);
                  fu2v2 = Proximity(u2, v2, j, r);
                  if((i!=mw && j!=mw && r!=mw && (fu1u2+fu1v2+fu2v2)/3>=threshold) ||
                     (i!=mw && j!=mw && r==mw && fu1u2>=threshold) ||
                     (i!=mw && j==mw && r!=mw && fu1v2>=threshold) ||
                     (i==mw && j!=mw && r!=mw && fu2v2>=threshold))
                     if((fu1v2+fu2v2) > maxsum)
                        { maxsum = fu1v2+fu2v2; maxpos2 = r; k2++; }
               }
               if(k2==0) {                   /* (virtual) L2 is empty */
                  *cuuij = 0;
                  suuij[0] = ZERO;           suuij[1] = ZERO;
                  suuij[2] = ZERO;           suuij[3] = ZERO;
               }
               else {                        /* take max on (virtual) L2 */
                  *cuuij = fu1u2 + maxsum;
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = ZERO;           suuij[3] = (short)maxpos2;
               }
               cuuij += nn4;  suuij += nn;      /* next j */
            }
      }
      else {                              /* Case 3: 2--2 */
         for(i=0; i<mwz; i++)
            for(j=0; j<mwz; j++) {
               fu1u2 = Proximity(u1, u2, i, j);
               k1=0;  q1=sborka->l1;                   /* Fill L1 */
               for(k=0; k<mwz; k++) {
                  fu1v1 = Proximity(u1, v1, i, k);
                  fu2v1 = Proximity(u2, v1, j, k);
                  if(i!=mw && j!=mw) {       /* Case a: u1!=0, u2!=0 */
                     if((k!=mw && fu1v1>=threshold && (fu1u2+fu2v1)/2>=threshold) ||
                        (k==mw && fu1u2>=threshold) ||
                        (k!=mw && fu1u2>=threshold && fu2v1>=threshold))
                           { *(q1++) = (short)k;  k1++; }
                  }
                  else if(i==mw && j!=mw) {  /* Case c: u1==0, u2!=0 */
                     if(k!=mw && fu2v1>=threshold)
                        { *(q1++) = (short)k;  k1++; }
                  }
                  else {                     /* Cases b,d: u2==0 */
                     if(k!=mw && fu1v1>=threshold)
                        { *(q1++) = (short)k;  k1++; }
                  }
               }
               k2=0;  q2=sborka->l2;                   /* Fill L2 */
               for(r=0; r<mwz; r++) {
                  fu1v2 = Proximity(u1, v2, i, r);
                  fu2v2 = Proximity(u2, v2, j, r);
                  if(i!=mw && j!=mw) {       /* Case a: u1!=0, u2!=0 */
                     if((r!=mw && fu2v2>=threshold && (fu1u2+fu1v2)/2>=threshold) ||
                        (r==mw && fu1u2>=threshold) ||
                        (r!=mw && fu1u2>=threshold && fu1v2>=threshold))
                           { *(q2++) = (short)r;  k2++; }
                  }
                  else if(i!=mw && j==mw) {  /* Case c: u1!=0, u2==0 */
                     if(r!=mw && fu1v2>=threshold)
                        { *(q2++) = (short)r;  k2++; }
                  }
                  else {                     /* Cases b,d: u1==0 */
                     if(r!=mw && fu2v2>=threshold)
                        { *(q2++) = (short)r;  k2++; }
                  }
               }
                                    /* Pair-wise exhaustion for v1,v2 */
               if(k1==0 && k2==0) {             /* Like in 1--1 case */
                  if(fu1u2 >= threshold) {
                     *cuuij = fu1u2;
                     suuij[0] = (short)i;    suuij[1] = (short)j;
                     suuij[2] = ZERO;        suuij[3] = ZERO;
                  }
                  else {
                     *cuuij = 0;
                     suuij[0] = ZERO;        suuij[1] = ZERO;
                     suuij[2] = ZERO;        suuij[3] = ZERO;
                  }
               }
               else if(k2==0) {                 /* Like in 2--1 case */
                  maxsum=-1;  q1=sborka->l1;
                  for(t1=0; t1<k1; t1++) {
                     k = *(q1++);
                     fu1v1 = Proximity(u1, v1, i, k);
                     fu2v1 = Proximity(u2, v1, j, k);
                     psum1 = fu1v1+fu2v1;
                     if(psum1 > maxsum)
                        { maxsum = psum1; maxpos1 = k; }
                  }
                  *cuuij = fu1u2 + maxsum;
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = (short)maxpos1; suuij[3] = ZERO;
               }
               else if(k1==0) {                 /* Like in 1--2 case */
                  maxsum=-1;  q2=sborka->l2;
                  for(t2=0; t2<k2; t2++) {
                     r = *(q2++);
                     fu1v2 = Proximity(u1, v2, i, r);
                     fu2v2 = Proximity(u2, v2, j, r);
                     psum2 = fu1v2+fu2v2;
                     if(psum2 > maxsum)
                        { maxsum = psum2; maxpos2 = r; }
                  }
                  *cuuij = fu1u2 + maxsum;
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = ZERO;           suuij[3] = (short)maxpos2;
               }
               else {            /* Regular case, pair-wise exhaustion */
                  maxsum=-1;  q1=sborka->l1;
                  for(t1=0; t1<k1; t1++) {
                     k = *(q1++);
                     fu1v1 = Proximity(u1, v1, i, k);
                     fu2v1 = Proximity(u2, v1, j, k);
                     psum1 = fu1u2+fu1v1+fu2v1;  q2=sborka->l2;
                     for(t2=0; t2<k2; t2++) {
                        r = *(q2++);
                        fu1v2 = Proximity(u1, v2, i, r);
                        fu2v2 = Proximity(u2, v2, j, r);
                        fv1v2 = Proximity(v1, v2, k, r);
                        psum2 = fu1v2+fu2v2+fv1v2;
                        if((psum1+psum2) > maxsum) {
                           maxsum = psum1+psum2;
                           maxpos1 = k;  maxpos2 = r;
                        }
                     }
                  }
                  *cuuij = maxsum;
                  suuij[0] = (short)i;       suuij[1] = (short)j;
                  suuij[2] = (short)maxpos1; suuij[3] = (short)maxpos2;
               }
               cuuij += nn4;  suuij += nn;      /* next j */
            }
      }
   }
   return rc;
}

