/* Regulatory Signal Search (RSS) */
/* Functions: MakeNextLevel */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"      /* Consider joint case of L1,2=0 & u1,2=0 ! */

int MakeNextLevel(SBORKA *sborka, int z) {
/* Builds suu[] system for k-th level of G-tree */
/*  based on suv[] system of the previous level. */
/* Argument z varies from logn-2 to 0 */
int rc=OK, sys_size, nn4, nn2, nnp, nnp2, fc_size;
int part, part1, node, u1, v1, u2, v2, v, s, t;
int i, j, k, r, k1, k2, t1, t2, maxpos1=-1, maxpos2=-1, maxsum, psum1, psum2, counter;
int fu1u2, fu1v2, fu2v1, fv1v2, off0z, offzz;
int *cuuij, *cuv1, *cuv2, *c1, *c2;
short *suuij, *q1, *q2, *tmp, *suv1, *suv2, *s1, *s2;
   if(!sborka) return ERMEM;
   sys_size = mwz*mwz*nn;
   nn4 = 1<<(z-1);   /* number of partitions at next level */
   nn2 = nn4 * 2;    /* number of partitions at previous level */
   nnp = nn / nn4;   /* number of nodes per partition at next level */
   nnp2 = nnp / 2;   /* number of nodes per partition at previous level */
   fc_size = mwz * mwz * nn4;
   off0z = mw*nn;          /* offset from <i,0> to <i,Z> in a system */
   offzz = mw*(mwz+1)*nn;  /* offset from <0,0> to <Z,Z> in a system */
                                    /* Perform step of induction */
   tmp = sborka->suv;  sborka->suv = sborka->suu;  sborka->suu = tmp;
   for(i=0,tmp=sborka->suu; i<sys_size; i++,tmp++)  *tmp = EMPTY;
   if(sborka->cuv)
      { free(sborka->cuv);  sborka->cuv = NULL; }
   sborka->cuv = sborka->cuu;
   if((sborka->cuu = (int*)calloc(fc_size, sizeof(int)))==NULL)
      return ERMEM;
   for(k=0,cuuij=sborka->cuu; k<fc_size; k++,cuuij++)  *cuuij=EMPTY;
                                       /* Loop by partitions */
   for(part1=0,part=0,node=0; node<nn; part1++,part+=2,node+=nnp) {
      u1 = sborka->tree[z][node];
      v1 = sborka->tree[z][node+1];
      u2 = sborka->tree[z][node+nnp2];
      v2 = sborka->tree[z][node+nnp2+1];
      cuuij = sborka->cuu + part1;  /* cu[i]u[j] (0,0) for current part. */
      suuij = sborka->suu + node;   /* su[i]u[j] (0,0) of the part. system */

      if(v1==EMPTY || v2==EMPTY)          /* Illegal cases 1, 2a, 2b */
         return ERINT;
      else {                              /* Case 3: nnp2 -- nnp2 */
         cuv1 = sborka->cuv+part;   /* cu[i]v[k] (0,0) for left subpartition */
         suv1 = sborka->suv+node;   /* su[i]v[k] (0,0) of left subpart.sys. */
         for(i=0; i<mwz; i++) {
            cuv2 = sborka->cuv+part+1;    /* cu[j]v[r] (0,0) for right subpart.*/
            suv2 = sborka->suv+node+nnp2; /* su[j]v[r] (0,0) of right sub.sys. */
            for(j=0; j<mwz; j++) {
               fu1u2 = Proximity(u1, u2, i, j);
               k1=0;  q1=sborka->l1;                   /* Fill L1 */
               c1 = cuv1;                 /* Goto cu[i]v[0] */
               s1 = suv1;                 /* Goto su[i]v[0] */
               for(k=0; k<mwz; k++) {
                  fu2v1 = Proximity(u2, v1, j, k);
                  if(i!=mw && j!=mw) {       /* Case a: u1!=0, u2!=0 */
                     if(*c1!=0) {      /* Calc. average prox. s(u1,v1)--u2 */
                        psum1=fu1u2;  counter=1;
                        for(s=1; s<nnp2; s++) {
                           if((v = sborka->tree[z][node+s])==EMPTY)
                              continue;
                           if((t = s1[s])==mw)   continue;
                           psum1 += Proximity(u2, v, j, t);
                           counter++;
                        }
                        if(psum1/counter>=threshold)
                           { *(q1++) = (short)k;  k1++; }
                     }
                     else if(fu1u2>=threshold && (k==mw || fu2v1>=threshold))
                        { *(q1++) = (short)k;  k1++; }
                  }
                  else if(i==mw && j!=mw) {  /* Case c: u1==0, u2!=0 */
                     if(*c1!=0) {      /* Calc. average prox. s(u1,v1)--u2 */
                        psum1=0;  counter=0;
                        for(s=1; s<nnp2; s++) {
                           if((v = sborka->tree[z][node+s])==EMPTY)
                              continue;
                           if((t = s1[s])==mw)   continue;
                           psum1 += Proximity(u2, v, j, t);
                           counter++;
                        }
                        if(psum1/counter>=threshold)
                           { *(q1++) = (short)k;  k1++; }
                     }
                     else if(fu2v1>=threshold)
                           { *(q1++) = (short)k;  k1++; }
                  }
                  else {                     /* Case b,d: u2==0 */
                     if(*c1!=0)
                        { *(q1++) = (short)k;  k1++; }
                  }
                  c1 += nn2;  s1 += nn;      /* next k */
               }
               k2=0;  q2=sborka->l2;                   /* Fill L2 */
               c2 = cuv2;                 /* Goto cu[j]v[0] */
               s2 = suv2;                 /* Goto su[j]v[0] */
               for(r=0; r<mwz; r++) {
                  fu1v2 = Proximity(u1, v2, i, r);
                  if(i!=mw && j!=mw) {       /* Case a: u1!=0, u2!=0 */
                     if(*c2!=0) {      /* Calc. average prox. s(u2,v2)--u1 */
                        psum2=fu1u2;  counter=1;
                        for(s=1; s<nnp2; s++) {
                           if((v = sborka->tree[z][node+nnp2+s])==EMPTY)
                              continue;
                           if((t = s2[s])==mw)   continue;
                           psum2 += Proximity(u1, v, i, t);
                           counter++;
                        }
                        if(psum2/counter>=threshold)
                           { *(q2++) = (short)r;  k2++; }
                     }
                     else if(fu1u2>=threshold && (r==mw || fu1v2>=threshold))
                        { *(q2++) = (short)r;  k2++; }
                  }
                  else if(i!=mw && j==mw) {  /* Case c: u1!=0, u2==0 */
                     if(*c2!=0) {      /* Calc. average prox. s(u2,v2)--u1 */
                        psum2=0;  counter=0;
                        for(s=1; s<nnp2; s++) {
                           if((v = sborka->tree[z][node+nnp2+s])==EMPTY)
                              continue;
                           if((t = s2[s])==mw)   continue;
                           psum2 += Proximity(u1, v, i, t);
                           counter++;
                        }
                        if(psum2/counter>=threshold)
                           { *(q2++) = (short)r;  k2++; }
                     }
                     else if(fu1v2>=threshold)
                           { *(q2++) = (short)r;  k2++; }
                  }
                  else {                     /* Case b,d: u1==0 */
                     if(*c2!=0)
                        { *(q2++) = (short)r;  k2++; }
                  }
                  c2 += nn2;  s2 += nn;      /* next r */
               }
                                    /* Pair-wise exhaustion for v1,v2 */
               if(k1==0 && k2==0) {             /* Like in 1--1 case */
                  if(fu1u2>=threshold) {
                     *cuuij = fu1u2;
                     suuij[0] = (short)i;    suuij[1] = (short)j;
                     suuij[2] = ZERO;        suuij[3] = ZERO;
                     s1 = suv1+off0z+2;      s2 = suv2+off0z+2;
                  }
                  else {
                     *cuuij = 0;
                     suuij[0] = ZERO;        suuij[1] = ZERO;
                     suuij[2] = ZERO;        suuij[3] = ZERO;
                     s1 = sborka->suv+node+offzz+2;
                     s2 = sborka->suv+node+nnp2+offzz+2;
                  }
                                       /* Merge tails of the subsystem */
                  for(k=4; k<nnp; k+=2)   suuij[k] = *(s1++);
                  for(k=5; k<nnp; k+=2)   suuij[k] = *(s2++);
               }
               else if(k2==0) {                 /* Like in 2--1 case */
                  maxsum=-1;  q1=sborka->l1;
                  for(t1=0; t1<k1; t1++) {
                     k = *(q1++);      /* Current v1 candidate position */
                     fu2v1 = Proximity(u2, v1, j, k);
                     psum1 = cuv1[k*nn2];     /* cu[i]v[k] */
                     psum1 += fu2v1;
                     if(psum1 > maxsum)
                        { maxsum = psum1; maxpos1 = k; }
                  }
                  if(maxsum > 0) {
                     *cuuij = fu1u2 + maxsum;
                     suuij[0] = (short)i;       suuij[1] = (short)j;
                     suuij[2] = (short)maxpos1; suuij[3] = ZERO;
                     s1 = suv1+maxpos1*nn+2;    s2 = suv2+off0z+2;
                  }
                  else {
                     *cuuij = 0;
                     suuij[0] = ZERO;           suuij[1] = ZERO;
                     suuij[2] = ZERO;           suuij[3] = ZERO;
                     s1 = sborka->suv+node+offzz+2;
                     s2 = sborka->suv+node+nnp2+offzz+2;
                  }
                                          /* Merge tails of the subsystems */
                  for(k=4; k<nnp; k+=2)   suuij[k] = *(s1++);
                  for(k=5; k<nnp; k+=2)   suuij[k] = *(s2++);
               }
               else if(k1==0) {                 /* Like in 1--2 case */
                  maxsum=-1;  q2=sborka->l2;
                  for(t2=0; t2<k2; t2++) {
                     r = *(q2++);      /* Current v2 candidate position */
                     fu1v2 = Proximity(u1, v2, i, r);
                     psum2 = cuv2[r*nn2];    /* cu[i]v[r] */
                     psum2 += fu1v2;
                     if(psum2 > maxsum)
                        { maxsum = psum2; maxpos2 = r; }
                  }
                  if(maxsum > 0) {
                     *cuuij = fu1u2 + maxsum;
                     suuij[0] = (short)i;       suuij[1] = (short)j;
                     suuij[2] = ZERO;           suuij[3] = (short)maxpos2;
                     s1 = suv1+off0z+2;         s2 = suv2+maxpos2*nn+2;
                  }
                  else {
                     *cuuij = 0;
                     suuij[0] = ZERO;           suuij[1] = ZERO;
                     suuij[2] = ZERO;           suuij[3] = ZERO;
                     s1 = sborka->suv+node+offzz+2;
                     s2 = sborka->suv+node+nnp2+offzz+2;
                  }
                                          /* Merge tails of the subsystems */
                  for(k=4; k<nnp; k+=2)   suuij[k] = *(s1++);
                  for(k=5; k<nnp; k+=2)   suuij[k] = *(s2++);
               }
               else {            /* Regular case, pair-wise exhaustion */
                  maxsum=-1;  q1=sborka->l1;
                  for(t1=0; t1<k1; t1++) {
                     k = *(q1++);      /* Current v1 candidate position */
                     fu2v1 = Proximity(u2, v1, j, k);
                     psum1 = cuv1[k*nn2];     /* cu[i]v[k] */
                     psum1 += fu1u2+fu2v1;  q2=sborka->l2;
                     for(t2=0; t2<k2; t2++) {
                        r = *(q2++);      /* Current v2 candidate position */
                        fu1v2 = Proximity(u1, v2, i, r);
                        fv1v2 = Proximity(v1, v2, k, r);
                        psum2 = cuv2[r*nn2];    /* cu[j]v[r] */
                        psum2 += fu1v2+fv1v2;
                        if((psum1+psum2) > maxsum) {
                           maxsum = psum1+psum2;
                           maxpos1 = k;  maxpos2 = r;
                        }
                     }
                  }
                  if(maxsum > 0) {
                     *cuuij = maxsum;
                     suuij[0] = (short)i;       suuij[1] = (short)j;
                     suuij[2] = (short)maxpos1; suuij[3] = (short)maxpos2;
                     s1 = suv1+maxpos1*nn+2;    s2 = suv2+maxpos2*nn+2;
                  }
                  else {
                     *cuuij = 0;
                     suuij[0] = ZERO;           suuij[1] = ZERO;
                     suuij[2] = ZERO;           suuij[3] = ZERO;
                     s1 = sborka->suv+node+offzz+2;
                     s2 = sborka->suv+node+nnp2+offzz+2;
                  }
                                          /* Merge tails of the subsystems */
                  for(k=4; k<nnp; k+=2)   suuij[k] = *(s1++);
                  for(k=5; k<nnp; k+=2)   suuij[k] = *(s2++);
               }
               cuuij += nn4;     suuij += nn;         /* next j */
               cuv2 += mwz*nn2;  suv2 += mwz*nn;
            }
            cuv1 += mwz*nn2;  suv1 += mwz*nn;   /* next i */
         }
      }
   }
   return rc;
}

