/* Regulatory Signal Search (RSS) */
/* Functions: Permutation */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include "rss01.h"

SBORKA* Permutation(int *rc) {
/* Allocates main structure for a new permutation and returns its pointer */
int i, j, k;
short *q;
SBORKA *newsborka;
   if((pmode==PM_IDENT && first>0) ||
      (pmode==PM_CYCLE && first>=n) ||
      first >= plimit) {
      if(cdef)
      	printf("Terminated due to chosen mode or limit of iterations\n");
      if(debug)
      	fprintf(outfile, "Terminated due to chosen mode or limit of iterations\n");
      return NULL;      /* Terminate by mode(s) and/or repetition limit */
   }
   if((newsborka = (SBORKA*)malloc(sizeof(SBORKA)))==NULL) {
      *rc |= ERMEM;
      return NULL;
   }
                                 /* Initialize the new sborka */
   newsborka->p = NULL;    newsborka->tree = NULL;
   newsborka->suu = NULL;  newsborka->suv = NULL;
   newsborka->l1 = NULL;   newsborka->l2 = NULL;
   newsborka->cuu = NULL;  newsborka->cuv = NULL;
   if((newsborka->p = (short*)calloc(n, sizeof(short)))==NULL) {
      *rc |= ERMEM;
      return NULL;
   }

   if(pmode==PM_IDENT || pmode==PM_CYCLE) /* Generate cyclic permutation(s) */
      for(i=0,j=first,q=newsborka->p; i<n; i++,j++) {
         if(j>=n) j=0;
         q[i] = (short)j;
      }
   else if(pmode==PM_RANDOM) {            /* Generate random permutation(s) */
      do {
         for(i=0,q=newsborka->p; i<n; i++,q++)  *q=EMPTY;
         for(i=0; i<NMAX; i++)  pvect[0][i]=EMPTY;
         for(i=0,q=newsborka->p; i<n; )
#ifdef MYRANDOM
            if(q[(j=MyRandom()%n)]==EMPTY)
#else
            if(q[(j=rand()%n)]==EMPTY)
#endif
               pvect[0][j] = q[j] = (short)(i++);
      } while(Recurrent(0));
   }
   else if(pmode==PM_ACTIVE) {   /* <=pvmax permutations may be buffered */
      if(first==0) {             /* Initially pvmax random permutations */
         for(k=0; k<pvmax; k++) {
            for(i=0; i<n; )
#ifdef MYRANDOM
               if(pvect[k][(j=MyRandom()%n)]==EMPTY)
#else
               if(pvect[k][(j=rand()%n)]==EMPTY)
#endif
                  pvect[k][j]=(short)i++;
            if(Recurrent(k)) {
               for(i=0; i<n; i++)  pvect[k][i]=EMPTY;
               k--;
            }
         }
			if(cdef>1)
         	printf("Initially generate %d random permutation(s)\n", pvmax);
         if(debug>1)
            fprintf(outfile, "Initially generate %d random permutation(s)\n", pvmax);
      }
      for(k=0; k<pvmax; k++) {         /* Find first unprocessed */
         if(pvect[k][0]==EMPTY)  continue;
         for(i=0,q=newsborka->p; i<n; i++,q++)
            *q = pvect[k][i];
         for(i=0; i<NMAX; i++)  pvect[k][i]=EMPTY;
         break;
      }
      if(k>=pvmax) {                   /* Complete queue processed - */
         if(AnalyzeHistory()!=OK)      /* generate new portion if possible */
            return NULL;
         for(k=0; k<pvmax; k++)        /* Take an element from the portion */
            if(pvect[k][0]!=EMPTY)  break;
         if(k>=pvmax)  return NULL;
         for(i=0,q=newsborka->p; i<n; i++,q++)
            *q = pvect[k][i];
         for(i=0; i<NMAX; i++)  pvect[k][i]=EMPTY;
      }
   }
   if(debug>1) {
      fprintf(outfile, "P[%d]: ", first);
      for(i=0,q=newsborka->p; i<n; i++,q++)
         fprintf(outfile, "%d ", *q+1);
      fprintf(outfile, "\n");
      if(fflush(outfile)!=0)  { *rc |= ERIO;  return newsborka; }
   }
#ifdef RSSDEBUG
      printf("P[%d]: ", first);
      for(i=0,q=newsborka->p; i<n; i++,q++)  printf("%d ", *q+1);
      printf("\n");
#endif
   first++;
   return newsborka;
}

