/* Regulatory Signal Search (RSS) */
/* Functions: MakeProximityMatrix, Proximity */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rss01.h"

int MakeProximityMatrix(void) {
/* Allocates and fills a matrix of pair-wise proximity for all words */
/*  of all sequences. */
int rc=OK, i, j, k, u, v, tcs, imax, jmax, prox;
char *pu, *pv;
BYTE *ptr;
   tcs = mw*mw;         /* Do not worry: n=n0 and nt=nt0 in this function */
   if((fuv = (BYTE**)calloc(nt, sizeof(BYTE*)))==NULL ||
      (ptr = (BYTE*)calloc(nt*tcs, sizeof(BYTE)))==NULL)
         return ERMEM;
   for(k=0; k<nt; k++,ptr+=tcs)
      fuv[k]=ptr;
   ptr = fuv[0];
   for(u=0; u<n-1; u++) {          /* Select sequence u */
      imax = seqw[u];
      for(v=u+1; v<n; v++) {              /* Select sequence v */
         jmax = seqw[v];
         for(i=0; i<imax; i++) {             /* Select next word from u */
            for(j=0; j<jmax; j++) {             /* Select next word from v */
               pu = seq[u]+i;
               pv = seq[v]+j;
               prox = 0;
               for(k=0; k<l; k++,pu++,pv++)        /* Calculate proximity */
                  if(*pu==*pv)  prox++;
               *(ptr++) = (BYTE)prox;              /* Store in matrix */
            }
            for(j=jmax; j<mw; j++)              /* Clear for tail of v if any */
               *(ptr++) = 0;
         }
         for(i=imax; i<mw; i++)              /* Clear for tail of u if any */
            for(j=0; j<mw; j++)
               *(ptr++) = 0;
      }
   }
   if(debug>2) {                /* Output of the proximity matrix */
      fprintf(outfile, "Proximity matrix: [u,v=%dx%d][i,j=%dx%d]\n",
         n, n, mw, mw);
      ptr=fuv[0];
      for(u=0; u<n-1; u++) {
         for(v=u+1; v<n; v++) {
            fprintf(outfile, "[%02d,%02d]", u, v);
            for(j=0; j<mw; j++)  fprintf(outfile, " %02d", j);
            fprintf(outfile, "\n");
            for(i=0; i<mw; i++) {
               fprintf(outfile, "   %02d: ", i);
               for(j=0; j<mw; j++)  fprintf(outfile, " %2d", *(ptr++));
               fprintf(outfile, "\n");
            }
            fprintf(outfile, "\n");
         }
      }
      if(fflush(outfile)!=0)  rc |= ERIO;
   }
   return rc;
}

short Proximity(int u, int v, int iu, int jv) {
/* Returns proximity (as number of coincidents) for two words of length l */
/* taken from the sequences with numbers u and v starting from positions iu */
/* and iv, respectively. Returns 0 if any word is zero-word, i.e. iu (or jv) */
/* equals mw. It does not check for end of sequence for quickness sake ! */
int k, prox;
char *pu, *pv;
BYTE *ptr;
   if(iu==mw || jv==mw)   return 0;
   if(matrix) {
   /*   if(iu>=seqw[u] || jv>=seqw[v]) return 0; */
      if(u>v) { k=u; u=v; v=k; k=iu; iu=jv; jv=k; }   /* Swap (u,v) if needed */
      k = u*(n0-1) - (u*(u+1))/2 + v-1;      /* u,v -> k (in triangle matrix) */
      ptr = fuv[k] + iu*mw + jv;
      return (short)(*ptr);
   }
   pu = seq[u]+iu;               /* Calculate proximity on the fly */
   pv = seq[v]+jv;
   prox = 0;
   for(k=0; k<l; k++,pu++,pv++)
      if(*pu==*pv)  prox++;
   return (short)prox;
}

