/* Regulatory Signal Search (RSS) */
/* Functions: ReadDataPortion */
/* L.Rubanov, 2006 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "rss01.h"

int ReadDataPortion(int *rc) {
/* Reads, checks and stores the next portion of input data */
/* Modifies *rc with ERIO, ERMEM if needed */
/* Returns only flags EODATA, ERDATA as appropriate */
int rc1=OK;
int len, i;
char *result, c, *q;
char sbuf[MMAX+3];     /* buffer to read sequence with CR+LF */

   if(!infile) {
      rc1 |= (EODATA | ERDATA);
      goto Final;
   }
   for(i=0; i<NMAX; i++)  seqw[i]=0;   /* Clear word counters */
   for(n=0,m=0; ;) {
      result = fgets(sbuf, MMAX+2, infile);
      if(result==NULL) {               /* EOF or I/O error */
         if(feof(infile)) {
            rc1 |= EODATA;
            if(n!=0 && (n<NMIN || n>NMAX))
               rc1 |= ERDATA;
         }
         else *rc |= ERIO;
         goto Final;
      }
      if((len = strlen(sbuf))>MMAX)    /* String too long */
         rc1 |= ERDATA;
      else {                           /* Remove new line character(s) */
         if(sbuf[len-1]=='\n' || sbuf[len-1]=='\r')
            sbuf[--len]=0;
         if(len>0 && (sbuf[len-1]=='\n' || sbuf[len-1]=='\r'))
            sbuf[--len]=0;
      }
      if(len==0) {    /* Empty line(s) = end of portion */
         if(n==0)  continue;
         if(n<NMIN || n>NMAX)    rc1 |= ERDATA;
         break;
      }
      if(strchr(COMMENTS, sbuf[0])) {  /* Comment string */
      	if((q=strchr(sbuf+1, ' '))!=NULL) *q=0;
         len=strlen(sbuf+1);
         if(len>MAXNAME)		rc1 |= ERDATA;
         if(len>maxname) maxname=len;
         if((q = malloc(len+1))==NULL) {	/* No memory */
         	*rc |= ERMEM;
            continue;				/* to reach end of portion/file */
         }
         strcpy(q, sbuf+1);
         name[n]=q;
         continue;
      }
                              /* Process data string */
      if(m < len)    m = len;
      if((q = malloc(len+1))==NULL) {   /* Not enough memory */
         *rc |= ERMEM;
         continue;            /* to reach end of portion/file */
      }
      seq[n]=q;
      for(i=0; i<len; i++) {  /* normalize and check */
         c = sbuf[i] = (char)tolower(sbuf[i]);
#ifdef RSSUTEQUIV
			if(c=='u')  c = sbuf[i] = 't';
#endif
         if(strchr(ALPHABET, c)==NULL)
            rc1 |= ERDATA;
      }
      seqw[n] = len-l+1;     /* number of words excluding zero-word */
      strcpy(seq[n++], sbuf);
   }
Final:   
   if(n<=0)  return rc1;
   if(n==2)  { nn=4; logn=2; }      /* NB: special case ! */
   else for(nn=1,logn=0; nn<n; nn*=2,logn++); /* Calc logn and 2**logn */
   nt = n*(n-1)/2;
   n0 = n;  nt0 = nt;  logn0 = logn;
   mw = m-l+1;          /* Maximum number of words excluding zero word */
   mwz = mw+1;          /* Maximum number of words including zero word */
   if(debug)
      fprintf(outfile, "\nn=%d  logn=%d  nn=%d  m<=%d  l=%d  mw<=%d(+1)"
         "  threshold=%d  seed=%d\n\n", n, logn, nn, m, l, mw, threshold, seed);
	if(cdef)
   	printf("Job accepted (n=%d  m<=%d  l=%d  s=%d)\n", n, m, l, seed);
   if(debug>1) {
      fprintf(outfile, "Sequences:\n");
      for(i=0; i<n; i++)
         fprintf(outfile, "%3d:%.*s %s\n", i+1, maxname, name[i], seq[i]);
      fprintf(outfile, "\n");
      if(fflush(outfile)!=0)  *rc |= ERIO;
   }
   return rc1;
}

