/* Regulatory Signal Search (RSS) */
/* Main function */
/* L.Rubanov, 2006 */
#define RSSDECLARE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "rss01.h"

int main(int argc, char **argv) {
int rc=OK, defpair=0, rc1;
int i, j, pvmax0;
SBORKA *sborka;
time_t t1, t2;
                     /* Initialize and set defaults */
   t1 = time(NULL);
   l = EMPTY;
   threshold = EMPTY;
   infile = outfile = (FILE*) NULL;
   fuv = NULL;
   if((seq = (char**)calloc(NMAX, sizeof(char*)))==NULL)
      { rc |= ERMEM;  goto Final; }
   if((name = (char**)calloc(NMAX, sizeof(char*)))==NULL)
      { rc |= ERMEM;  goto Final; }
   for(i=0; i<NMAX; i++)
   	{ seq[i]=NULL; name[i]=NULL; }
   debug=DG_DEF;  pmode=PM_DEF;
   plimit=LIM_DEF;  aggr=AG_DEF;
   aftermax=ITER_DEF;  afterabs=ITER_DEF*10;
   pvmax=PV_DEF;  pairs=EMPTY;  matrix=1;
   rdef=RDEF;  cdef=CDEF;  seed=(unsigned)t1;
   printf("\n(C) 2006, Laboratory of Mathematical Methods and Models in Bioinformatics\n"
          "Institute for Information Transmission Problems, Russian Academy of Sciences\n"
   		 "Program for Regulatory Signal Search (single CPU, v.%s)    L.Rubanov, 2006\n\n", VERSION);
                     /* Analyze command line arguments, if any */
   if(argc > 1)
      if((rc |= AnalyzeArgs(argc, argv)) != OK) goto Final;
   pvmax0 = pvmax;
   if(pairs==EMPTY)  defpair=1;
   if(pmode==PM_FULL) {
      fprintf(stderr, "Permutation mode %d has not been implemented yet\n", pmode);
      rc |= ERARG;
      goto Final;
   }
                     /* Read and check input data portions */
   while(1) {
      for(i=0; i<NMAX; i++) {       /* Free data memory allocated earlier */
         if(seq[i])  { free(seq[i]); seq[i] = NULL; }
         if(name[i]) { free(name[i]); name[i]=NULL; } 
      }
      maxname=0;
      if(fuv) {                     /* Free proximity matrix */
         if(*fuv)  free(*fuv);
         free(fuv);  fuv=NULL;
      }
      FreeHistory();                /* Free previous history list if any */
      rc1=ReadDataPortion(&rc);
      if(n==0 && (rc1 & EODATA))
         break;                     /* Exit data loop */
      if((rc1 & ERDATA) | (rc & ERMEM)) {
         rc |= ERDATA;
         continue;                  /* Skip portion with errors */
      }

      if(matrix && MakeProximityMatrix()==ERMEM) {
         matrix = 0;    /* No memory for matrix - calculate on the fly */
			if(cdef>1)
         	printf("No memory for matrix - calculate on the fly\n");
      }
      else if(matrix && cdef>1)
         printf("Proximity matrix built: [triangle %d*%d/2][%d*%d]\n",
            (n-1), n, mw, mw);
      else if(cdef>1)
         printf("Proximity is calculated on the fly\n");
                                    /* Allocates profile and coverage */
      if((lambda = (int**)calloc(n, sizeof(int*)))==NULL ||
         (cover = (int*)calloc(nt, sizeof(int)))==NULL)
         { rc |= ERMEM; goto Final; }
      for(i=0; i<n; i++) {
         if((lambda[i] = (int*)calloc(mw, sizeof(int)))==NULL)
            { rc |= ERMEM; goto Final; }
      }
      for(i=0; i<nt; i++)  cover[i]=0;
      first = 0;           /* For pmode==PM_IDENT||PM_CYCLE||PM_RANDOM */
#ifdef MYRANDOM
		MySrand(seed);
#else
      srand(seed);
#endif
      recursum = 0;        /* Clear counter of recurrent permutations */
      after = 0;
      if(defpair)    pairs = n/2;
      for(i=0; i<=MAXPVECT; i++)
         for(j=0; j<NMAX; j++)
            pvect[i][j]=EMPTY;
      if(n<5 && pvmax0>n)  pvmax=n;    /* Limit pvmax for small n */ 
      else pvmax=pvmax0;
      /* -- Permutation loop: multiple repetition of the following loop may be
      		performed in parallel, provided write access to the common memory
            of all repetitions (i.e. history, lambda, cover...) is controlled */
      while((sborka=Permutation(&rc)) != NULL) {
         rc |= GenTree(sborka);
         rc |= Construction(sborka);
         rc |= AddHistory(sborka);
         FreeSborka(sborka);
         if(rc & ERMEM) {
            fprintf(stderr, "Not enough memory or invalid request\n");
            break;
         }
      }
      /* ------------------ Wait here for all parallel repetitions to complete */
      if(rc & ~ERDATA) break;    /* Finish at any error except ERDATA */

      for(i=1; i<=rdef; i++)				/* Select & print rdef best result(s) */
         if((rc1=BestResult(i))!=OK) {
            rc |= rc1;
            fprintf(stderr, "Error %d creating result\n", rc1);
            break;
         }
      if(debug) {
         fprintf(outfile, "Recurrent permutations rejected: %d\n\n", recursum);
         if(fflush(outfile)!=0) rc |= ERIO;
      }
		if(cdef>1)
      	printf("Recurrent permutations rejected: %d\n\n", recursum);
   }
Final:
   t2 = time(NULL);
   if(!(rc & HELP)) {
      if(outfile && debug)
         fprintf(outfile, "RC: %d  (time: %-.0f sec)\n\n", rc, difftime(t2,t1));
      printf("RC:");
      if (rc==OK) printf(" OK");
      else if(rc & ERARG)  printf(" ERARG");
      else if(rc & ERDATA) printf(" ERDATA");
      else if(rc & ERIO)   printf(" ERIO");
      else if(rc & EODATA) printf(" EODATA");
      else if(rc & ERMEM)  printf(" ERMEM");
      else if(rc & EDEBUG) printf(" EDEBUG");
      else if(rc & ERINT)  printf(" ERINT");
      printf("  (time: %-.0f sec)\n\n", difftime(t2,t1));
   }
   for(i=0; i<NMAX; i++) {      /* Free data memory allocated earlier */
      if(seq[i])  { free(seq[i]); seq[i] = NULL; }
      if(name[i]) { free(name[i]); name[i] = NULL; }
   }
   FreeHistory();
   if(infile && fclose(infile)==EOF)    rc |= ERIO;
   if(outfile && fclose(outfile)==EOF)   rc |= ERIO;
	printf("Thank you for using the program.\n\n");
   return rc;
}

